import numpy as np
from clease.montecarlo.averager import Averager
from .mc_observer import MCObserver


class SGCObserver(MCObserver):
    """
    Observer mainly intended to track additional quantities needed when
    running SGC Monte Carlo. This observer has to be executed on every MC step.

    Parameters:

    calc: `clease.calculators.Clease`
        Clease calculator
    """

    name = "SGCObersver"

    def __init__(self, calc):
        super().__init__()
        self.calc = calc
        E = calc.get_potential_energy()
        n_singlets = len(self.calc.get_singlets())
        self.quantities = {
            "singlets": np.zeros(n_singlets, dtype=np.float64),
            "singlets_sq": np.zeros(n_singlets, dtype=np.float64),
            "energy": Averager(ref_value=E),
            "energy_sq": Averager(ref_value=E**2),
            "singl_eng": np.zeros(n_singlets, dtype=np.float64),
            "counter": 0
        }

    def reset(self):
        """Reset all variables to zero."""
        self.quantities["singlets"][:] = 0.0
        self.quantities["singlets_sq"][:] = 0.0
        self.quantities["energy"].clear()
        self.quantities["energy_sq"].clear()
        self.quantities["singl_eng"][:] = 0.0
        self.quantities["counter"] = 0

    def __call__(self, system_changes):
        """Update all SGC parameters.

        Parameters:

        system_changes: list
            System changes. See doc-string of
            `clease.montecarlo.observers.MCObserver`
        """
        self.quantities["counter"] += 1
        new_singlets = self.calc.get_singlets()

        E = self.calc.results['energy']

        self.quantities["singlets"] += new_singlets
        self.quantities["singlets_sq"] += new_singlets**2
        self.quantities["energy"] += E
        self.quantities["energy_sq"] += E**2
        self.quantities["singl_eng"] += new_singlets * E

    @property
    def energy(self):
        return self.quantities["energy"]

    @property
    def energy_sq(self):
        return self.quantities["energy_sq"]

    @property
    def singlets(self):
        return self.quantities["singlets"]

    @property
    def singl_eng(self):
        return self.quantities["singl_eng"]

    @property
    def counter(self):
        return self.quantities["counter"]

    def interval_ok(self, interval):
        return interval == 1



import copy
from collections import Counter
from itertools import product
from typing import NamedTuple, Tuple, List
import numpy as np
from scipy.spatial.distance import cdist
from scipy.optimize import linear_sum_assignment
from ase.atoms import Cell
from ase import Atoms
import attr

__all__ = ('FourVector', 'construct_four_vectors')


def _int_converter(x):
    """Allow for NumPy integer types (convert to regular int)"""
    if isinstance(x, np.integer):
        return int(x)
    return x


@attr.s(frozen=True, order=True)
class FourVector:
    """Container for a vector in 4-vector space, i.e. a vector of [ix, iy, iz, sublattice].
    Represents the position of a site in terms of the number of repitions of the primitive atoms,
    as well as which sublattice it belongs to in that cell.
    """
    ix: int = attr.ib(converter=_int_converter, validator=attr.validators.instance_of(int))
    iy: int = attr.ib(converter=_int_converter, validator=attr.validators.instance_of(int))
    iz: int = attr.ib(converter=_int_converter, validator=attr.validators.instance_of(int))
    sublattice: int = attr.ib(converter=_int_converter, validator=attr.validators.instance_of(int))

    def to_cartesian(self, prim: Atoms, transposed_cell: np.ndarray = None) -> np.ndarray:
        """Convert the four vector into cartesian coordinates

        Args:
            prim (Atoms): Primitive Atoms object that defines the lattice

            transposed_cell (np.ndarray, optional): The transposed cell matrix.
             Can be passed to avoid re-calculating this property, e.g. if getting
             cartesian coordinates for multiple FourVector objects simultaniously.
             If transposed_cell is None, it is calculated from the primitive atoms object.
             Defaults to None.

        Returns:
            np.ndarray: The cartesian coordinate representation of the FourVector.
        """
        if transposed_cell is None:
            transposed_cell = prim.get_cell().T
        return np.dot(transposed_cell, self.xyz_array) + prim[self.sublattice].position

    def to_scaled(self, prim: Atoms) -> np.ndarray:
        """
        Convert the four vector into scaled coordinates

        :param prim: Primitive cell that defines the lattice
        """
        return self.xyz_array + prim.get_scaled_positions()[self.sublattice, :]

    @property
    def xyz_array(self) -> np.ndarray:
        """Return the [ix, iy, iz] component of the four-vector as a NumPy array."""
        return np.array([self.ix, self.iy, self.iz])

    def to_tuple(self) -> Tuple[int]:
        """Get the tuple representation of the four-vector
        """
        return attr.astuple(self)

    def copy(self) -> 'FourVector':
        """Create a copy of the FourVector instance."""
        return copy.copy(self)

    def shift_xyz(self, other: 'FourVector') -> 'FourVector':
        """Shift a this vector by another FourVector instance.
        Only translates the x, y, and z values, the sublattice remains the original,
        *so be mindful of the order*, i.e. a.shift_xyz(b) is not the same as b.shift_xyz(a).

        Example:

        >>> from clease.datastructures.four_vector import FourVector
        >>> a = FourVector(0, 0, 0, 0)
        >>> b = FourVector(1, 0, 0, 1)
        >>> a.shift_xyz(b)
        FourVector(ix=1, iy=0, iz=0, sublattice=0)
        >>> b.shift_xyz(a)
        FourVector(ix=1, iy=0, iz=0, sublattice=1)
        """
        if not isinstance(other, FourVector):
            raise NotImplementedError(f'Shift must be by another FourVector, got {type(other)}')

        return FourVector(self.ix + other.ix, self.iy + other.iy, self.iz + other.iz,
                          self.sublattice)


class _Box(NamedTuple):
    """
    Class defining a box. The box is defined by storing the position vector of
    the start (lower) and end (upper) of the diagonal. The names lower and upper
    refers to "lower corner" and "upper corner", respectively.
    """
    lower: np.ndarray
    upper: np.ndarray


def _supercell_is_integer_multiple(prim: Cell, supercell: Cell, tol: float = 1e-8) -> bool:
    """
    Return True ff C_{super} = C_{prim}P and P is an integer matrix. The columns of C_{super}
    is the cell vectors of the supercell and the columns of C_{prim} are the cell vectors of
    the primitive cell

    :param prim: Primitive cell
    :param supercell: Supercell
    :param tol: Tolerance for comparison
    """
    P = np.linalg.solve(prim.array.T, supercell.array.T)
    P_int = np.round(P)
    return np.max(np.abs(P - P_int)) < tol


def _make_grid(bbox: _Box, prim: Atoms) -> Tuple[List[FourVector], np.ndarray]:
    """
    Make a grid bounded by bbox using repetitions of the atomic positions in prim

    :param bbox: Bounding box for the grid
    :param prim: Primitive cell that defines the sublattice

    :return:
        four_vecs: List of FourVector of length N (where N is the number of grid
                   points)
        scaled_pos: Nx3 numpy array. Each row of scaled_pos corresponds to the
            four-vector in four_vecs
    """
    four_vecs = []
    scaled_pos = []

    ranges = [range(bbox.lower[i], bbox.upper[i] + 1) for i in range(3)] + [range(len(prim))]
    for ix, iy, iz, l in product(*ranges):
        fv = FourVector(ix, iy, iz, l)
        four_vecs.append(fv)
        scaled_pos.append(fv.to_scaled(prim))

    # Convert to 2D numpy array and perform a sanity check on the result
    scaled_pos = np.array(scaled_pos)
    assert scaled_pos.shape == (len(four_vecs), 3)
    return four_vecs, scaled_pos


def construct_four_vectors(prim: Atoms, supercell: Atoms) -> List[FourVector]:
    """
    Calculate the all four vector corresponding to the passed Atoms object. A four-vector u is
    defined by such that the position of an atom is given by C.dot(u[:3]) + prim[u[3]].position
    where the cell vectors of the primitive cell are column in the C matrix.

    The cell of the supercell must be related to the cell in prim my an integer matrix
    transformation. C_{super} = C_{prim}P. See for example `make_supercell`in ASE.

    :param prim: Primitive cell
    :param supercell: Supercell
    """
    if not _supercell_is_integer_multiple(prim.get_cell(), supercell.get_cell()):
        raise ValueError("The supercell must be an integer multiple of the primitive cell")

    if np.any(prim.get_scaled_positions() < 0.0):
        raise ValueError("Primitive cell must be wrapped")

    supercell_cpy = supercell.copy()
    supercell_cpy.set_cell(prim.get_cell())
    sc_scaled_pos = supercell_cpy.get_scaled_positions(wrap=False)

    bounding_box = _Box(lower=np.floor(np.min(sc_scaled_pos, axis=0)).astype(int),
                        upper=np.ceil(np.max(sc_scaled_pos, axis=0)).astype(int))

    # Fill the bounding box with a grid
    four_vecs_grid, scaled_pos_grid = _make_grid(bounding_box, prim)

    # Calculates pair-wise distances between the super-cell and the grid
    distances = cdist(sc_scaled_pos, scaled_pos_grid)

    # Assign exactly one atom to each grid point such that the total distance mismatch is minimize
    # (e.g. solving least linear sym assignment problem)
    _, cols = linear_sum_assignment(distances)

    # Calculate mean displacement
    four_vectors = [four_vecs_grid[c] for c in cols]

    # Sanity checks
    sub_lat_count = Counter(fv.sublattice for fv in four_vectors)

    if any(v != int(len(supercell_cpy) / len(prim)) for v in sub_lat_count.values()):
        raise RuntimeError("The four vectors are invalid as the number of atoms in each "
                           "sublattice does not match the number of primitive cells")

    if len(sub_lat_count) != len(prim):
        raise RuntimeError("The four vectors are invalid as the number of sublattice "
                           "detected does not match the number of sublattices in the "
                           "primitive cell")

    return four_vectors



"""Definitions of Cluster Expansion settings for bulk.

Cluster Expansion can be peformed on bulk material using either CEBulk
or CECrystal class.
"""
from copy import deepcopy

from ase.build import bulk
from ase.spacegroup import crystal
from clease.tools import wrap_and_sort_by_position
from .concentration import Concentration
from .settings import ClusterExpansionSettings

__all__ = (
    'CEBulk',
    'CECrystal',
)


def CEBulk(concentration: Concentration,
           crystalstructure='sc',
           a=None,
           c=None,
           covera=None,
           u=None,
           **kwargs):
    """
    Specify cluster expansion settings for bulk materials defined based on
    crystal structures.

    Parameters:

        concentration (Union[Concentration, dict]):
            Concentration object or dictionary specifying the basis elements and
            concentration range of constituting species

        crystalstructure (str):
            Must be one of sc, fcc, bcc, hcp, diamond, zincblende, rocksalt,
            cesiumchloride, fluorite or wurtzite.

        a (float):
            Lattice constant.

        c (float):
            Lattice constant.

        covera (float):
            c/a ratio used for hcp.  Default is ideal ratio: sqrt(8/3).

        u (float):
            Internal coordinate for Wurtzite structure.

    For more kwargs, see docstring of :class:`clease.settings.ClusterExpansionSettings`.
    """
    structures = {
        'sc': 1,
        'fcc': 1,
        'bcc': 1,
        'hcp': 1,
        'diamond': 1,
        'zincblende': 2,
        'rocksalt': 2,
        'cesiumchloride': 2,
        'fluorite': 3,
        'wurtzite': 2
    }

    num_basis = len(concentration.orig_basis_elements)
    if num_basis != structures[crystalstructure]:
        msg = f"{crystalstructure} has {structures[crystalstructure]} basis. "
        msg += "The number of basis specified by basis_elements is "
        msg += f"{num_basis}"
        raise ValueError(msg)

    basis_elements = concentration.orig_basis_elements
    name = ''.join(x[0] for x in basis_elements)
    prim = bulk(name=name, crystalstructure=crystalstructure, a=a, c=c, covera=covera, u=u)
    prim = wrap_and_sort_by_position(prim)

    settings = ClusterExpansionSettings(prim, concentration, **kwargs)

    settings.kwargs.update({
        'crystalstructure': crystalstructure,
        'a': a,
        'c': c,
        'covera': covera,
        'u': u,
        'factory': 'CEBulk'
    })
    return settings


def CECrystal(concentration: Concentration,
              spacegroup=1,
              basis=None,
              cell=None,
              cellpar=None,
              ab_normal=(0, 0, 1),
              crystal_kwargs=None,
              **kwargs):
    """Store CE settings on bulk materials defined based on space group.

    Parameters:

        concentration (Union[Concentration, dict]):
            Concentration object or dictionary specifying the basis elements and
            concentration range of constituting species

        spacegroup (int | string | Spacegroup instance):
            Space group given either as its number in International Tables
            or as its Hermann-Mauguin symbol.

        basis (List[float]):
            List of scaled coordinates.
            Positions of the unique sites corresponding to symbols given
            either as scaled positions or through an atoms instance.

        cell (3x3 matrix):
            Unit cell vectors.

        cellpar ([a, b, c, alpha, beta, gamma]):
            Cell parameters with angles in degree. Is not used when `cell`
            is given.

        ab_normal (vector):
            Is used to define the orientation of the unit cell relative
            to the Cartesian system when `cell` is not given. It is the
            normal vector of the plane spanned by a and b.

        crystal_kwargs (dict | None):
            Extra kwargs to be passed into the ase.spacegroup.crystal
            function. Nothing additional is added if None.
            Defaults to None.

    For more kwargs, see docstring of :class:`clease.settings.ClusterExpansionSettings`.
    """

    symbols = []
    num_basis = len(concentration.orig_basis_elements)
    for x in range(num_basis):
        symbols.append(concentration.orig_basis_elements[x][0])

    crystal_kwargs = crystal_kwargs or {}
    prim = crystal(symbols=symbols,
                   basis=basis,
                   spacegroup=spacegroup,
                   cell=cell,
                   cellpar=cellpar,
                   ab_normal=ab_normal,
                   size=[1, 1, 1],
                   primitive_cell=True,
                   **crystal_kwargs)
    prim = wrap_and_sort_by_position(prim)

    settings = ClusterExpansionSettings(prim, concentration, **kwargs)
    settings.kwargs.update({
        'basis': deepcopy(basis),
        'spacegroup': spacegroup,
        'cell': cell,
        'cellpar': cellpar,
        'ab_normal': ab_normal,
        'factory': 'CECrystal'
    })
    return settings



# pylint: disable=undefined-variable
# For some reason, pylint detects a lot of cyclic imports here.
from .concentration import *
from . import template_filters
from . import template_atoms
from . import atoms_manager
from .settings import *
from .settings_bulk import *
from .settings_slab import *
from .utils import *

from . import settings

ADDITIONAL = ('settings', 'template_atoms', 'template_filters', 'atoms_manager')
__all__ = (concentration.__all__ + settings.__all__ + settings_bulk.__all__ +
           settings_slab.__all__ + utils.__all__ + ADDITIONAL)



from typing import Sequence, Set, Dict, List
from itertools import product
from copy import deepcopy
import numpy as np
import ase
from ase.geometry import wrap_positions

from clease.datastructures import FourVector, Figure

from .cluster_fingerprint import ClusterFingerprint
from .cluster import Cluster
from .cluster_list import ClusterList
from .cluster_generator import ClusterGenerator
from .utils import name_clusters, size

__all__ = ('ClusterManager',)


class ClusterManager:
    """
    Manager for construction of all cluster.

    Parameters:

    prim_cell: ase.Atoms
        Primitive cell
    """

    def __init__(self, prim_cell: ase.Atoms, background_syms: Set[str] = None):
        self._background_syms = background_syms or set()

        primitive_filtered = self._filter_background(prim_cell)

        self.generator = ClusterGenerator(primitive_filtered)
        self.clusters = ClusterList()
        self._cache = _CacheChecker()

    @property
    def prim(self) -> ase.Atoms:
        """The primitive cell.
        Note, that background atoms have been removed from this atoms object."""
        return self.generator.prim

    @property
    def background_syms(self) -> Set[str]:
        """The symbols which are considered background."""
        return self._background_syms

    def _filter_background(self, atoms: ase.Atoms) -> ase.Atoms:
        """Filter the background atoms from an ASE Atoms object.

        Returns a copy of the original atoms object.
        """
        atoms = atoms.copy()

        # Find the indices we need to delete
        # If no background atoms are present, this will do nothing.
        delete = [atom.index for atom in atoms if self.is_background_atom(atom)]
        delete.sort(reverse=True)
        for i in delete:
            del atoms[i]
        return atoms

    def is_background_atom(self, atom: ase.Atom) -> bool:
        """Check whether an atom is a background atom."""
        return atom.symbol in self.background_syms

    def __eq__(self, other):
        return self.clusters == other.clusters and \
            self.generator == other.generator

    def _prepare_new_build(self, max_cluster_dia):
        """Prepare for a new call to ``build``
        """
        # Update the cache
        self._cache.set_cache(max_cluster_dia)
        # Clear any old clusters
        self.clusters.clear()

    def requires_build(self, max_cluster_dia: Sequence[float]) -> bool:
        """Check if the cluster manager requires a new build
        for a given set of cluster diameters"""
        return self._cache.requires_build(max_cluster_dia)

    def build(self, max_cluster_dia: Sequence[float]) -> None:
        """
        Construct all clusters.

        Parameters:

        max_cluster_dia: sequence of floats
            Maximum distance between two atoms in a cluster,
            for each cluster body. The diameters start at 2-body clusters
        """
        # Check if we already built the clusters with these settings
        if not self.requires_build(max_cluster_dia):
            return
        # We got a new set of settings, prepare to construct new clusters
        self._prepare_new_build(max_cluster_dia)

        # Number of lattices from 0 to N, where N
        # is the number of atoms in the primitive,
        # possibly without the background atoms.
        num_lattices = range(len(self.prim))
        all_fps: List[ClusterFingerprint] = []
        all_figures: List[List[Figure]] = []
        lattices = []

        for ref_lattice, (indx, diameter) in product(num_lattices, enumerate(max_cluster_dia)):
            cluster_size = indx + 2  # Size of cluster, start with 2-body at index 0
            figures, fps = self.generator.generate(cluster_size, diameter, ref_lattice)

            all_fps += fps
            all_figures += figures
            lattices += [ref_lattice] * len(figures)

        names = self._get_names(all_fps)
        # Transfer to the cluster list
        for figures, fp, name, ref_lattice in zip(all_figures, all_fps, names, lattices):
            # All figures are of the same size
            cluster_size = figures[0].size
            # Calculate the diameter from the first Figure, since they are all
            # geometrically equivalent, and thus have the same diameter.
            diameter = figures[0].get_diameter(self.prim)
            eq_sites = self.generator.equivalent_sites(figures[0])

            cluster = Cluster(name=name,
                              size=cluster_size,
                              diameter=diameter,
                              fingerprint=fp,
                              figures=figures,
                              equiv_sites=eq_sites,
                              group=ref_lattice)
            self.clusters.append(cluster)

        # Add singlets
        for i in range(len(self.prim)):
            self.clusters.append(
                Cluster(name='c1',
                        size=1,
                        diameter=0.0,
                        fingerprint=ClusterFingerprint([1.0]),
                        figures=[Figure([FourVector(0, 0, 0, i)])],
                        equiv_sites=[],
                        group=i))
        # Add empty
        self.clusters.append(
            Cluster(name='c0',
                    size=0,
                    diameter=0.0,
                    fingerprint=ClusterFingerprint([0.0]),
                    figures=[],
                    equiv_sites=[],
                    group=0))
        # Put the clusters in order of size. Has no practical effect,
        # but it looks nicer upon inspection.
        self.clusters.sort()

    @staticmethod
    def _get_names(all_fps):
        """
        Give a consistent name to all clusters

        Parameter:

        all_fps: list of ClusterFingerPrint
            A list with all the cluster fingerprints
        """
        # The following one-liner will do the job when we
        # decide to update cluster names from the legacy ones
        # where names where per by size
        # names = name_clusters(all_fps)

        sizes = set(size(fp) for fp in all_fps)
        names = [None for _ in all_fps]
        for s in sizes:
            fps = []
            indices = []
            for i, fp in enumerate(all_fps):
                if size(fp) == s:
                    indices.append(i)
                    fps.append(fp)
            names_per_size = name_clusters(fps)
            for i, n in zip(indices, names_per_size):
                names[i] = n
        return names

    def info_for_template(self, template: ase.Atoms) -> ClusterList:
        """
        Specialise the cluster information to a template

        Parameter:

        template: ase.Atoms
            Atoms object representing the simulation cell
        """
        unique = self.unique_four_vectors()
        lut = self.fourvec_to_indx(template, unique)
        ref_indices = [lut[FourVector(0, 0, 0, i)] for i in range(self.generator.num_sub_lattices)]

        cluster_int = deepcopy(self.clusters)
        for cluster in cluster_int:
            if cluster.size == 0:
                cluster.ref_indx = int(ref_indices[cluster.group])
                cluster.indices = []
            elif cluster.size == 1:
                cluster.ref_indx = int(ref_indices[cluster.group])
                cluster.indices = []
            else:
                cluster.indices = self.generator.to_atom_index(cluster, lut)
                cluster.ref_indx = int(ref_indices[cluster.group])
        return cluster_int

    def unique_four_vectors(self) -> Set[FourVector]:
        """
        Return a list with all unique 4-vectors which are
        represented in any figure in all of the clusters.
        """
        # We utilize that FourVector objects are hashable,
        # and therefore can be filtered using a set()
        unique = set()
        for cluster in self.clusters:
            if cluster.size == 0:
                continue
            for figure in cluster.figures:
                for fv in figure.components:
                    unique.add(fv)
        return unique

    def get_figures(self) -> List[ase.Atoms]:
        """
        Return a list of atoms object representing the clusters
        """
        return self.clusters.get_figures(self.generator)

    def create_four_vector_lut(self, template: ase.Atoms) -> Dict[FourVector, int]:
        """
        Construct a lookup table (LUT) for the index in template given the
        wrapped vector

        Parameter:

        template: Atoms
            Atoms object to use when creating the lookup table (LUT)
        """
        lut = {}
        pos = template.get_positions().copy()
        for i in range(pos.shape[0]):
            if self.is_background_atom(template[i]):
                # No need to make a lookup for a background atom
                continue
            vec = self.generator.to_four_vector(pos[i, :], template[i].tag)
            lut[vec] = i
        return lut

    def fourvec_to_indx(self, template: ase.Atoms,
                        unique: Sequence[FourVector]) -> Dict[FourVector, int]:
        """Translate a set of unique FourVectors into their corresponding index
        in a template atoms object."""
        cell = template.get_cell()
        pos = np.zeros((len(unique), 3))
        for i, fv in enumerate(unique):
            pos[i, :] = fv.to_cartesian(self.prim)

        pos = wrap_positions(pos, cell)
        unique_indices = []
        for i in range(pos.shape[0]):
            diff_sq = np.sum((pos[i, :] - template.get_positions())**2, axis=1)
            unique_indices.append(np.argmin(diff_sq))
        return dict(zip(unique, unique_indices))

    def translation_matrix(self, template: ase.Atoms) -> List[Dict[int, int]]:
        """
        Construct the translation matrix.

        The translation matrix translates a given atomic index to
        the corresponding atomic site if we started from index 0.

        Parameter:

        template: ase.Atoms
            Atoms object representing the simulation cell
        """
        trans_mat = []  # The final translation matrix list
        cell = template.get_cell()

        # Get the unique four-vectors which are present in all of our clusters.
        unique = self.unique_four_vectors()

        lut = self.create_four_vector_lut(template)
        cartesian = np.zeros((len(unique), 3))

        # Map the un-translated unique 4-vectors to their index
        unique_indx_lut = self.fourvec_to_indx(template, unique)
        # call int() to convert from NumPy integer to python integer
        unique_index = [int(unique_indx_lut[u]) for u in unique]

        for atom in template:
            if self.is_background_atom(atom):
                # This atom is considered a background, it has no mapping
                trans_mat.append({})
                continue

            # Translate the atom into its four-vector representation
            vec = self.generator.to_four_vector(atom.position, sublattice=atom.tag)

            # Translate the (x, y, z) components of the unique four-vectors
            # by this atom's (x, y, z) four-vector component
            translated_unique = [u.shift_xyz(vec) for u in unique]

            # Find the new Cartesian coordinates of the translated FourVectors,
            # and wrap them back into the cell
            cartesian = self.generator.to_cartesian(*translated_unique)
            cartesian = wrap_positions(cartesian, cell)

            # Re-translate the wrapped-Cartesian coordinates of the unique four-vectors
            # into a four-vector representation (with a generator expression)
            four_vecs = (self.generator.to_four_vector(cart, fv.sublattice)
                         for cart, fv in zip(cartesian, translated_unique))

            # Get the index of the translated four-vector
            indices = [lut[fv] for fv in four_vecs]

            trans_mat.append(dict(zip(unique_index, indices)))
        return trans_mat


class _CacheChecker:
    """Helper class to check if the
    cluster manager has already been built
    using a given set of settings.
    """

    def __init__(self):
        self.max_cluster_dia = None

    def requires_build(self, max_cluster_dia: Sequence[float]) -> bool:
        """Check if a given set of 'max_size'
        and 'max_cluster_dia' has previously been
        used to build the clusters"""
        # We don't have anything cached yet
        if self.max_cluster_dia is None:
            return True

        # Check if the parameters match
        return not np.array_equal(self.max_cluster_dia, max_cluster_dia)

    def set_cache(self, max_cluster_dia: Sequence[float]):
        # Ensure we set a copy, so no external
        # mutations affect the cache
        self.max_cluster_dia = deepcopy(max_cluster_dia)



"""Module for defining deprecated regression class imports"""
# pylint: skip-file
from deprecated import deprecated
from clease import regression

__all__ = ('LinearRegression', 'Tikhonov', 'Lasso', 'BayesianCompressiveSensing',
           'ConstrainedRidge', 'GAFit', 'GeneralizedRidgeRegression', 'PhysicalRidge',
           'SequentialClusterRidge')

MSG = 'Import {} from clease.regression instead'
DEP_VERSION = '0.10.0'  # Deprecation version


@deprecated(version=DEP_VERSION, reason=MSG.format('LinearRegression'))
class LinearRegression(regression.LinearRegression):
    pass


@deprecated(version=DEP_VERSION, reason=MSG.format('Tikhonov'))
class Tikhonov(regression.Tikhonov):
    pass


@deprecated(version=DEP_VERSION, reason=MSG.format('Lasso'))
class Lasso(regression.Lasso):
    pass


@deprecated(version=DEP_VERSION, reason=MSG.format('BayesianCompressiveSensing'))
class BayesianCompressiveSensing(regression.BayesianCompressiveSensing):
    pass


@deprecated(version=DEP_VERSION, reason=MSG.format('ConstrainedRidge'))
class ConstrainedRidge(regression.ConstrainedRidge):
    pass


@deprecated(version=DEP_VERSION, reason=MSG.format('GAFit'))
class GAFit(regression.GAFit):
    pass


@deprecated(version=DEP_VERSION, reason=MSG.format('GeneralizedRidgeRegression'))
class GeneralizedRidgeRegression(regression.GeneralizedRidgeRegression):
    pass


@deprecated(version=DEP_VERSION, reason=MSG.format('PhysicalRidge'))
class PhysicalRidge(regression.PhysicalRidge):
    pass


@deprecated(version=DEP_VERSION, reason=MSG.format('SequentialClusterRidge'))
class SequentialClusterRidge(regression.SequentialClusterRidge):
    pass



from typing import Sequence, Dict
import numpy as np
from ase import Atoms
from ase.units import kB
from .observers import SGCObserver
from .montecarlo import Montecarlo
from .trial_move_generator import TrialMoveGenerator, RandomFlip


class InvalidChemicalPotentialError(Exception):
    pass


# pylint: disable=too-many-instance-attributes
class SGCMonteCarlo(Montecarlo):
    """
    Class for running Monte Carlo in the Semi-Grand Canonical Ensebmle
    (i.e., fixed number of atoms, but varying composition)

    See the docstring of :class:`clease.montecarlo.Montecarlo`

    :param atoms: Atoms object (with CLEASE calculator attached!)
    :param temp: Temperature in kelvin
    :param symbols: Possible symbols to be used in swaps
    :param generator: Generator that produces trial moves
    """

    def __init__(self,
                 atoms: Atoms,
                 temp: float,
                 symbols: Sequence[str] = (),
                 generator: TrialMoveGenerator = None):
        if generator is None:
            if len(symbols) <= 1:
                raise ValueError("At least 2 symbols have to be specified")
            generator = RandomFlip(symbols, atoms)
        super().__init__(atoms, temp, generator=generator)
        self.symbols = symbols

        self.averager = SGCObserver(self.atoms.calc)

        self.chem_pots = []
        self.chem_pot_names = []
        self.has_attached_avg = False
        self.name = "SGCMonteCarlo"
        self._chemical_potential = None
        self.chem_pot_in_eci = False
        self.current_singlets = None

        has_attached_obs = False
        for obs in self.observers:
            if obs.name == "SGCObserver":
                has_attached_obs = True
                self.averager = obs
                break
        if not has_attached_obs:
            self.attach(self.averager)

    def _check_symbols(self):
        """
        Override because there are no restriction on the symbols here
        """

    def reset(self):
        """
        Reset the simulation object
        """
        super().reset()
        self.averager.reset()

    @property
    def chemical_potential(self):
        return self._chemical_potential

    @chemical_potential.setter
    def chemical_potential(self, chem_pot: Dict[str, float]):
        eci = self.atoms.calc.eci
        if any(key not in eci for key in chem_pot):
            msg = "A chemical potential not being trackted is added. Make "
            msg += "sure that all the following keys are in the ECIs before "
            msg += "they are passed to the calculator: "
            msg += f"{list(chem_pot.keys())}\n"
            msg += "(Add them with a zero ECI value if they are not supposed "
            msg += "to be included.)"
            raise InvalidChemicalPotentialError(msg)

        self._chemical_potential = chem_pot
        if self.chem_pot_in_eci:
            self._reset_eci_to_original(self.atoms.calc.eci)
        self._include_chemical_potential_in_eci(chem_pot, self.atoms.calc.eci)

    def _include_chemical_potential_in_eci(self, chem_pot: Dict[str, float], eci: Dict[str, float]):
        """
        Including the chemical potentials in the ECIs

        Parameters:


        chem_pot: dict
            Chemical potentials
        eci: dict
            Original ECIs
        """
        self.chem_pots = []
        self.chem_pot_names = []
        keys = list(chem_pot.keys())
        keys.sort()
        for key in keys:
            self.chem_pots.append(chem_pot[key])
            self.chem_pot_names.append(key)
            current_eci = eci.get(key, 0.0)
            eci[key] = current_eci - chem_pot[key]
        calc = self.atoms.calc
        calc.update_eci(eci)
        self.chem_pot_in_eci = True
        self.current_energy = calc.calculate(None, None, None)
        return eci

    def _reset_eci_to_original(self, eci_with_chem_pot: Dict[str, float]):
        """
        Resets the ECIs to their original value

        :parma dict eci_with_chem_pot: ECIs with chemical potential included
        """
        for name, val in zip(self.chem_pot_names, self.chem_pots):
            eci_with_chem_pot[name] += val
        calc = self.atoms.calc
        calc.update_eci(eci_with_chem_pot)
        self.chem_pot_in_eci = False
        self.current_energy = calc.calculate(None, None, None)
        return eci_with_chem_pot

    def reset_eci(self):
        """Return the ECIs."""
        if self.chem_pot_in_eci:
            self._reset_eci_to_original(self.atoms.calc.eci)

    # pylint: disable=arguments-differ
    def run(self, steps: int = 10, chem_pot: Dict[str, float] = None):
        """
        Run Monte Carlo simulation.
        See :py:meth:`cemc.mcmc.Montecarlo.runMC`

        Parameters:

        chem_pot: dict
            Chemical potentials. The keys should correspond to one of the
            singlet terms. A typical form of this is {"c1_0":-1.0,c1_1_1.0}
        """

        if chem_pot is None and self.chemical_potential is None:
            ex_chem_pot = {"c1_1": -0.1, "c1_2": 0.05}
            raise ValueError(f"No chemicalpotentials given. Has to be "
                             f"dictionary of the form {ex_chem_pot}")

        if chem_pot is not None:
            self.chemical_potential = chem_pot
        self.reset()

        Montecarlo.run(self, steps=steps)

    def singlet2composition(self, avg_singlets: Dict[str, float]):
        """Convert singlets to composition."""
        bf = self.atoms.calc.settings.basis_functions
        matrix = np.zeros((len(self.symbols), len(self.symbols)))

        index = {s: i for i, s in enumerate(self.symbols)}
        for i, b in enumerate(bf):
            for s, col in index.items():
                matrix[i, col] = b[s]

        matrix[-1, :] = 1.0
        rhs = np.zeros(len(self.symbols))
        rhs[:-1] = avg_singlets
        rhs[-1] = 1.0
        x = np.linalg.solve(matrix, rhs)

        res = {}
        for s, i in index.items():
            name = s + "_conc"
            res[name] = x[i]
        return res

    # pylint: disable=arguments-differ
    def get_thermodynamic_quantities(self, reset_eci=True):
        """Compute thermodynamic quantities.

        Parameters:

        reset_eci: bool
            If True, the chemical potential will be removed from the ECIs
        """
        N = self.averager.counter
        quantities = {}
        singlets = self.averager.singlets / N
        singlets_sq = self.averager.quantities["singlets_sq"] / N

        quantities["sgc_energy"] = self.averager.energy.mean
        quantities["sgc_heat_capacity"] = self.averager.energy_sq.mean - \
            self.averager.energy.mean**2

        quantities["sgc_heat_capacity"] /= (kB * self.T**2)

        quantities["energy"] = self.averager.energy.mean
        natoms = len(self.atoms)
        for i, chem_pot in enumerate(self.chem_pots):
            quantities["energy"] += chem_pot * singlets[i] * natoms

        quantities["temperature"] = self.T
        quantities["n_mc_steps"] = self.averager.counter

        # Add singlets and chemical potential to the dictionary
        # pylint: disable=consider-using-enumerate
        for i in range(len(singlets)):
            name = f"singlet_{self.chem_pot_names[i]}"
            quantities[name] = singlets[i]

            name = f"var_singlet_{self.chem_pot_names[i]}"
            quantities[name] = singlets_sq[i] - singlets[i]**2

            name = f"mu_{self.chem_pot_names[i]}"
            quantities[name] = self.chem_pots[i]

        quantities.update(self.meta_info)

        try:
            avg_conc = self.singlet2composition(singlets)
            quantities.update(avg_conc)
        # pylint: disable=broad-except
        except Exception as exc:
            print("Could not find average singlets!")
            print(exc)

        if reset_eci:
            self._reset_eci_to_original(self.atoms.calc.eci)
        return quantities



from pathlib import Path

with Path(__file__).with_name('_version.txt').open('r') as f:
    __version__ = f.readline().strip()

# Representation of version_info as (x, y, z)
version_info = __version__.split('.')

__all__ = ('__version__', 'version_info')
